# Configuration file for the Sphinx documentation builder.

import os
import sys

# -- Project information -----------------------------------------------------

project = 'EDL'
copyright = '2020-2024, Matthias Klumpp'
author = 'Matthias Klumpp'

# The full version, including alpha/beta/rc tags
release = '0.2'

# -- General configuration ---------------------------------------------------
thisfile = __file__
if not os.path.isabs(thisfile):
    thisfile = os.path.normpath(os.path.join(os.getcwd(), thisfile))
sys.path.insert(0, os.path.normpath(os.path.join(os.path.dirname(thisfile), '..')))

html_theme = 'pydata_sphinx_theme'
html_theme_options = {
    'show_nav_level': 2,
    'collapse_navigation': True,
    'github_url': 'https://github.com/bothlab/edlio',
}

html_css_files = [
    'custom.css',
]

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx_autodoc_typehints',
]

# Napoleon settings
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# Intersphinx
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'pint': ('https://pint.readthedocs.io/en/latest', None),
    'neo': ('https://neo.readthedocs.io/en/stable', None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['css']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import typing as T

import pytest


@pytest.fixture(scope='session')
def samples_dir() -> T.Iterator[str]:
    """
    Prepare samples in the samples directory and return full path to the directory.
    """

    import lzma

    from . import source_root

    smpldir = os.path.join(source_root, 'tests', 'samples')
    if not os.path.isdir(smpldir):
        raise RuntimeError('Unable to find test samples directory in {}'.format(smpldir))

    # unpack any xz files that we just compress for efficient storage,
    # but need in their raw form for testing
    raw_fnames = [os.path.join(smpldir, 'blink1', 'intan-signals', 'a870_data_210208_181726.rhd')]
    for raw_fname in raw_fnames:
        with open(raw_fname, 'wb') as f:
            f.write(lzma.open(raw_fname + '.xz').read())

    yield smpldir

    # cleanup
    for raw_fname in raw_fnames:
        os.remove(raw_fname)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import json
import struct
import typing as T
import logging as log
from enum import IntEnum
from uuid import UUID
from datetime import datetime, timezone

import pint
import numpy as np
from xxhash import xxh3_64

from .. import ureg

__all__ = ['TSyncFile', 'TSyncFileMode', 'TSyncTimeUnit']


class TSyncFileMode(IntEnum):
    """Time storage mode of a TSync file."""

    CONTINUOUS = 0  # Continous time-point mapping with no gaps
    SYNCPOINTS = 1  # Only synchronization points are saved


class TSyncTimeUnit(IntEnum):
    """Unit types for time representation in a TSync file."""

    INDEX = 0
    NANOSECONDS = 1
    MICROSECONDS = 2
    MILLISECONDS = 3
    SECONDS = 4


class TSyncDataType(IntEnum):
    """Data types use for storing time values in the data file."""

    INVALID = 0
    INT16 = 2
    INT32 = 3
    INT64 = 4

    UINT16 = 6
    UINT32 = 7
    UINT64 = 8


TSYNC_MAGIC = int('F223434E5953548A', 16)

TSYNC_VERSION_MAJOR = 1
TSYNC_VERSION_MINOR = 2

TSYNC_BLOCK_TERM = int('1126000000000000', 16)
TSYNC_BLOCK_TERM_32 = int('11260000', 16)


def tsync_dtype_to_pack_fmt_len(dtype: TSyncDataType) -> tuple[str, int]:
    """Convert tsync data type into Python unpack format string and length"""
    if dtype == TSyncDataType.INT16:
        return '<h', 2
    if dtype == TSyncDataType.UINT16:
        return '<H', 2
    if dtype == TSyncDataType.INT32:
        return '<i', 4
    if dtype == TSyncDataType.UINT32:
        return '<I', 4
    if dtype == TSyncDataType.INT64:
        return '<q', 8
    if dtype == TSyncDataType.UINT64:
        return '<Q', 8
    raise RuntimeError('No data defined for how to unpack type {}'.format(dtype))


def tsync_time_unit_to_punit(unit: TSyncTimeUnit) -> pint.Unit:
    if unit == TSyncTimeUnit.INDEX:
        return T.cast(pint.Unit, ureg.dimensionless)
    if unit == TSyncTimeUnit.NANOSECONDS:
        return T.cast(pint.Unit, ureg.nsec)
    if unit == TSyncTimeUnit.MICROSECONDS:
        return T.cast(pint.Unit, ureg.usec)
    if unit == TSyncTimeUnit.MILLISECONDS:
        return T.cast(pint.Unit, ureg.msec)
    if unit == TSyncTimeUnit.SECONDS:
        return T.cast(pint.Unit, ureg.sec)
    raise ValueError('Can not convert tsync time unit type "{}" to Pint unit type.'.format(unit))


def read_utf8_xxh_from_file(f: T.BinaryIO, xxh: T.Any) -> str:
    """Read UTF-8 encoded string from binary .tsync file"""

    (length,) = struct.unpack('<I', f.read(4))
    if length == int('ffffffff', 16):
        return ''

    if length > (os.fstat(f.fileno()).st_size - f.tell() + 1):
        raise ValueError('String length in binary too long ({}).'.format(length))

    data = f.read(length)
    xxh.update(data)
    return str(data, 'utf-8')


class TSyncFile:
    """
    Read a TimeSync (.tsync) binary file as generated by the
    Syntalos DAQ system.
    """

    def __init__(self, fname: str | None = None):
        self._format_version = '1.0'
        self._time_created: datetime | None = None
        self._generator_name = ''
        self._collection_id = UUID(int=0x00)
        self._ts_mode = TSyncFileMode.CONTINUOUS
        self._block_size = 128
        self._custom: dict[str, T.Any] = {}
        self._time_labels = ('A', 'B')
        self._time_units = (
            tsync_time_unit_to_punit(TSyncTimeUnit.MICROSECONDS),
            tsync_time_unit_to_punit(TSyncTimeUnit.MICROSECONDS),
        )
        self._times = np.empty((0, 2))
        if fname:
            self.open(fname)

    @property
    def time_created(self) -> datetime | None:
        return self._time_created

    @property
    def tolerance(self) -> int:
        """The tolerance range value, in microseconds"""
        return int(self._custom.get('tolerance_us', 0))

    @tolerance.setter
    def tolerance(self, usec: int) -> None:
        self._custom['tolerance_us'] = usec

    @property
    def generator_name(self) -> str:
        """Name of the module that generated this file."""
        return self._generator_name

    @generator_name.setter
    def generator_name(self, name: str) -> None:
        self._generator_name = name

    @property
    def collection_id(self) -> UUID:
        """Data collection ID this file belongs to."""
        return self._collection_id

    @collection_id.setter
    def collection_id(self, uuid: UUID) -> None:
        self._collection_id = uuid

    @property
    def sync_mode(self) -> TSyncFileMode:
        """Time data storage mode.."""
        return self._ts_mode

    @sync_mode.setter
    def sync_mode(self, mode: TSyncFileMode) -> None:
        self._ts_mode = mode

    @property
    def custom(self) -> dict[str, T.Any]:
        """User-defined custom properties of this file."""
        return self._custom

    @custom.setter
    def custom(self, v: dict[str, T.Any]) -> None:
        self._custom = v

    @property
    def time_labels(self) -> tuple[str, str]:
        """Labels of the two encoded times."""
        return self._time_labels

    @time_labels.setter
    def time_labels(self, v: tuple[str, str]) -> None:
        self._time_labels = v

    @property
    def time_units(self) -> tuple[T.Any, T.Any]:
        """Units of the two encoded times."""
        return self._time_units

    @time_units.setter
    def time_units(self, v: tuple[T.Any, T.Any]) -> None:
        self._time_units = v

    @property
    def times(self) -> np.ndarray:
        """The actual time values of the two clocks."""
        return self._times

    @times.setter
    def times(self, v: np.ndarray) -> None:
        self._times = v

    def _read_xxh_unpack(self, format: str, buffer: bytes) -> int:
        self._xxh.update(buffer)
        (v,) = struct.unpack(format, buffer)
        return int(v)

    def _read_utf8_xxh_from_file(self, f: T.BinaryIO) -> str:
        return read_utf8_xxh_from_file(f, self._xxh)

    def open(self, fname: str) -> None:
        with open(fname, 'rb') as f:
            (magic_number,) = struct.unpack('<Q', f.read(8))
            if magic_number != TSYNC_MAGIC:
                raise ValueError('Unrecognized file type: This file is no tsync file.')

            # read file header block
            self._xxh = xxh3_64()
            major_version = self._read_xxh_unpack('<H', f.read(2))
            minor_version = self._read_xxh_unpack('<H', f.read(2))
            self._format_version = '{}.{}'.format(major_version, minor_version)
            if major_version != TSYNC_VERSION_MAJOR or minor_version > TSYNC_VERSION_MINOR:
                raise ValueError(
                    'Can not read TSync format version {} (max {}.{})'.format(
                        self._format_version, TSYNC_VERSION_MAJOR, TSYNC_VERSION_MINOR
                    )
                )
            log.debug('Reading tsync {} file: {}'.format(self._format_version, fname))
            check_xxh = major_version >= 1 and minor_version >= 2
            if not check_xxh:
                log.warning(
                    'Tsync file version ({}) is too old, checksum validation for '
                    'integrity checks will be skipped.'.format(self._format_version)
                )

            self._time_created = datetime.fromtimestamp(
                self._read_xxh_unpack('<q', f.read(8)), tz=timezone.utc
            )
            self._generator_name = self._read_utf8_xxh_from_file(f)
            self._collection_id = UUID(self._read_utf8_xxh_from_file(f))
            user_json_raw = self._read_utf8_xxh_from_file(f)
            self._custom = {}
            if user_json_raw:
                self._custom = json.loads(user_json_raw)

            self._ts_mode = TSyncFileMode(self._read_xxh_unpack('<H', f.read(2)))
            self._block_size = self._read_xxh_unpack('<i', f.read(4))

            time1Name = self._read_utf8_xxh_from_file(f)
            time1Unit = TSyncTimeUnit(self._read_xxh_unpack('<H', f.read(2)))
            time1DType = TSyncDataType(self._read_xxh_unpack('<H', f.read(2)))

            time2Name = self._read_utf8_xxh_from_file(f)
            time2Unit = TSyncTimeUnit(self._read_xxh_unpack('<H', f.read(2)))
            time2DType = TSyncDataType(self._read_xxh_unpack('<H', f.read(2)))

            self._time_labels = (time1Name, time2Name)
            self._time_units = (
                tsync_time_unit_to_punit(time1Unit),
                tsync_time_unit_to_punit(time2Unit),
            )

            # skip alignment padding
            padding = (f.tell() * -1) & (8 - 1)
            self._xxh.update(f.read(padding))

            # check header CRC
            if check_xxh:
                term_bytecount = 16
                (block_term,) = struct.unpack('<Q', f.read(8))
                (expected_header_cs,) = struct.unpack('<Q', f.read(8))
                if block_term != TSYNC_BLOCK_TERM:
                    raise ValueError(
                        'Header block terminator not found: The file is either '
                        'invalid or its header block was damaged.'
                    )
                if expected_header_cs != self._xxh.intdigest():
                    raise ValueError(
                        'Header checksum mismatch: The file is either invalid or '
                        'its header block was damaged.'
                    )
            else:
                term_bytecount = 8
                (block_term,) = struct.unpack('<I', f.read(4))
                f.read(4)
                if block_term != TSYNC_BLOCK_TERM_32:
                    # check if we maybe had no padding due to an erroneous writer
                    f.seek((padding + 4 + 4) * -1, os.SEEK_CUR)
                    (block_term,) = struct.unpack('<I', f.read(4))
                    f.read(4)
                    if block_term != TSYNC_BLOCK_TERM_32:
                        raise ValueError(
                            'Header block terminator not found: The file is either '
                            'invalid or its header block was damaged.'
                        )

            self._xxh.reset()

            tfmt1, tlen1 = tsync_dtype_to_pack_fmt_len(time1DType)
            tfmt2, tlen2 = tsync_dtype_to_pack_fmt_len(time2DType)

            self._times = np.empty((0, 2))
            bytes_per_entry = tlen1 + tlen2
            bytes_per_block = bytes_per_entry * self._block_size + term_bytecount

            bytes_remaining = os.fstat(f.fileno()).st_size - f.tell()
            if bytes_remaining <= 0:
                # no data is present
                return

            whole_block_count = bytes_remaining // bytes_per_block
            last_block_bytes_remaining = bytes_remaining - (whole_block_count * bytes_per_block)
            if last_block_bytes_remaining == 0:
                last_block_len = 0
            else:
                last_block_len_f = (last_block_bytes_remaining - term_bytecount) / bytes_per_entry
                if last_block_len_f.is_integer() and last_block_len_f > 0:
                    last_block_len = int(last_block_len_f)
                else:
                    raise ValueError(
                        'File "{}" may be corrupt: Suspicious size ({}) of '
                        'last data block.'.format(fname, last_block_len_f)
                    )
            entries_n = whole_block_count * self._block_size + last_block_len

            self._times = np.zeros((entries_n, 2), dtype=np.int64)
            self._block_crc = 0
            b_index = 0
            i = 0
            while True:
                if bytes_remaining == 0:
                    break

                time1 = self._read_xxh_unpack(tfmt1, f.read(tlen1))
                time2 = self._read_xxh_unpack(tfmt2, f.read(tlen2))
                bytes_remaining -= bytes_per_entry
                self._times[i] = np.array([time1, time2])

                i += 1
                b_index += 1
                if b_index == self._block_size or bytes_remaining == term_bytecount:
                    bytes_remaining -= term_bytecount

                    if not check_xxh:
                        (block_term,) = struct.unpack('<I', f.read(4))
                        f.read(4)
                        if block_term != TSYNC_BLOCK_TERM_32:
                            raise ValueError(
                                'Block terminator not found: Some data may be corrupted.'
                            )
                        b_index = 0
                        continue

                    # check validity of the block we read last
                    (block_term,) = struct.unpack('<Q', f.read(8))
                    (expected_cs,) = struct.unpack('<Q', f.read(8))
                    if block_term != TSYNC_BLOCK_TERM:
                        raise ValueError(
                            'Block terminator not found: Some data is likely corrupted.'
                        )
                    if expected_cs != self._xxh.intdigest():
                        raise ValueError('Block checksum mismatch: Some data is likely corrupted.')
                    self._xxh.reset()
                    b_index = 0
            del self._xxh


class LegacyTSyncFile:
    """
    Read a legacy TimeSync (.tsync) binary file as generated by the
    Syntalos DAQ system (legacy variant for an older, experimental
    version of this file format that was briefly in use).
    """

    def __init__(self, fname: str | None = None):
        self._format_version = 1
        self._time_created: datetime | None = None
        self._tolerance_us = 0
        self._generator_name = ''
        self._custom: dict[str, T.Any] = {}
        self._time_labels = ('A', 'B')
        self._time_units = (
            tsync_time_unit_to_punit(TSyncTimeUnit.MICROSECONDS),
            tsync_time_unit_to_punit(TSyncTimeUnit.MICROSECONDS),
        )
        self._times = np.empty((0, 2))
        if fname:
            self.open(fname)

    @property
    def time_created(self) -> datetime | None:
        return self._time_created

    @property
    def tolerance(self) -> int:
        """The tolerance range value, in microseconds"""
        return self._tolerance_us

    @tolerance.setter
    def tolerance(self, usec: int) -> None:
        self._tolerance_us = usec

    @property
    def generator_name(self) -> str:
        """Name of the module that generated this file."""
        return self._generator_name

    @generator_name.setter
    def generator_name(self, name: str) -> None:
        self._generator_name = name

    @property
    def sync_mode(self) -> TSyncFileMode:
        """Time data storage mode.."""
        return TSyncFileMode.SYNCPOINTS

    @property
    def custom(self) -> dict[str, T.Any]:
        """User-defined custom properties of this file."""
        return self._custom

    @custom.setter
    def custom(self, v: dict[str, T.Any]) -> None:
        self._custom = v

    @property
    def time_labels(self) -> tuple[str, str]:
        """Labels of the two encoded times."""
        return self._time_labels

    @time_labels.setter
    def time_labels(self, v: tuple[str, str]) -> None:
        self._time_labels = v

    @property
    def time_units(self) -> tuple[T.Any, T.Any]:
        """Units of the two encoded times."""
        return self._time_units

    @time_units.setter
    def time_units(self, v: tuple[T.Any, T.Any]) -> None:
        self._time_units = v

    @property
    def times(self) -> np.ndarray:
        """The actual time values of the two clocks."""
        return self._times

    @times.setter
    def times(self, v: np.ndarray) -> None:
        self._times = v

    @staticmethod
    def is_legacy(fname: str) -> bool:
        with open(fname, 'rb') as f:
            (magic_number,) = struct.unpack('<I', f.read(4))
            return bool(magic_number == int('C6BBDFBC', 16))

    def open(self, fname: str) -> None:
        with open(fname, 'rb') as f:
            (magic_number,) = struct.unpack('<I', f.read(4))
            if magic_number != int('C6BBDFBC', 16):
                raise ValueError('Unrecognized file type.')

            (self._format_version,) = struct.unpack('<I', f.read(4))
            if self._format_version != 1:
                raise ValueError(
                    'Can not read TSync format version {}'.format(self._format_version)
                )
            log.debug('Reading legacy tsync file: {}'.format(fname))

            (ts,) = struct.unpack('<q', f.read(8))
            self._time_created = datetime.fromtimestamp(ts, tz=timezone.utc)
            (self._tolerance_us,) = struct.unpack('<I', f.read(4))

            xxh = xxh3_64()
            try:
                self._generator_name = read_utf8_xxh_from_file(f, xxh)
            except UnicodeDecodeError as uni_e:
                raise ValueError(
                    'This legacy tsync file is damaged and can not be read.'
                ) from uni_e

            json_raw = read_utf8_xxh_from_file(f, xxh)
            self._custom = {}
            if json_raw:
                self._custom = json.loads(json_raw)

            tlabel1 = read_utf8_xxh_from_file(f, xxh)
            tlabel2 = read_utf8_xxh_from_file(f, xxh)
            self._time_labels = (tlabel1, tlabel2)

            (tuv1,) = struct.unpack('<H', f.read(2))
            (tuv2,) = struct.unpack('<H', f.read(2))
            self._time_units = (
                tsync_time_unit_to_punit(TSyncTimeUnit(tuv1)),
                tsync_time_unit_to_punit(TSyncTimeUnit(tuv2)),
            )

            self._times = np.empty((0, 2))
            bytes_per_block = 4 + 8 + 8
            bytes_remaining = os.fstat(f.fileno()).st_size - f.tell()
            if bytes_remaining <= 0:
                # no data is present
                return

            if bytes_remaining % bytes_per_block != 0:
                raise ValueError('File may be corrupt: Not a whole number of data blocks found!')

            num_data_blocks = int(bytes_remaining / bytes_per_block)
            indices_continuous = True
            self._times = np.zeros((num_data_blocks, 2), dtype=np.int64)
            for i in range(num_data_blocks):
                (index,) = struct.unpack('<I', f.read(4))
                (time1,) = struct.unpack('<q', f.read(8))
                (time2,) = struct.unpack('<q', f.read(8))
                if index != i:
                    indices_continuous = False
                self._times[i] = np.array([time1, time2])

            if not indices_continuous:
                print('WARNING: Indices in time sync file were not continuous.')


def load_data(part_paths: T.Iterable[str], aux_data_list: T.Any) -> T.Iterator[T.Any]:
    """Entry point for automatic dataset loading.

    This function is used internally to Syntalos' .tsync files
    as data or auxiliary data.
    """
    for fname in part_paths:
        tsync: LegacyTSyncFile | TSyncFile
        if LegacyTSyncFile.is_legacy(fname):
            tsync = LegacyTSyncFile(fname)
        else:
            tsync = TSyncFile(fname)
        yield tsync



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os
import uuid
import typing as T
from datetime import datetime

import tomlkit as toml

from .utils import listify, sanitize_name

# version of the EDL specification
EDL_FORMAT_VERSION: str = '1'


def make_collection_uuid() -> uuid.UUID:
    """Create an UUID for use in an EDL collection."""
    f = getattr(uuid, 'uuid7', None)
    if f is not None:
        return f()
    return uuid.uuid4()


class EDLError(Exception):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message: str):
        self.message = message
        super(EDLError, self).__init__(self.message)


class EDLUnit:
    """
    Generic base class for all EDL unit types.
    """

    def __init__(self, name: str | None = None):
        self._parent: EDLUnit | None = None
        self._name = sanitize_name(name)
        self._collection_id = make_collection_uuid()
        self._root_path: str | None = None
        self._authors: list[T.Any] = []
        self._attrs: dict[str, T.Any] = {}
        self._format_version = EDL_FORMAT_VERSION
        self._generator_id: str | None = None
        self._unit_type = self._type_as_unittype()
        self._time_created = datetime.now().replace(microsecond=0)

    @property
    def parent(self) -> EDLUnit | None:
        return self._parent

    @property
    def name(self) -> str | None:
        return self._name

    @property
    def unit_type(self) -> str:
        return self._unit_type

    @property
    def time_created(self) -> datetime:
        return self._time_created

    @time_created.setter
    def time_created(self, v: datetime) -> None:
        self._time_created = v

    @property
    def collection_id(self) -> uuid.UUID:
        return self._collection_id

    @collection_id.setter
    def collection_id(self, v: uuid.UUID) -> None:
        self._collection_id = v

    @property
    def path(self) -> str:
        if not self._root_path or not self._name:
            raise RuntimeError('Path of this unit was empty.')
        return os.path.join(self._root_path, self._name)

    @property
    def root_path(self) -> str | None:
        return self._root_path

    @root_path.setter
    def root_path(self, path: str) -> None:
        self._root_path = path

    @property
    def authors(self) -> list[T.Any]:
        return self._authors

    @property
    def attributes(self) -> dict:
        return self._attrs

    @attributes.setter
    def attributes(self, v: dict[str, T.Any]) -> None:
        self._attrs = v

    def change_name(self, new_name: str) -> None:
        old_dir_path = self.path
        old_name = self.name

        # set new name, but sanitize it to be portable across many filesystems
        self._name = sanitize_name(new_name)

        if old_dir_path and os.path.exists(old_dir_path):
            try:
                if not self.path:
                    raise RuntimeError('New path was empty.')
                os.rename(old_dir_path, self.path)
            except Exception as e:
                self._name = old_name
                raise ValueError('Unable to set new unit name: {}'.format(str(e))) from e

    def load(self, path: str | os.PathLike, mf: T.MutableMapping[str, T.Any] | None = None) -> None:
        """
        Load an EDL unit from a path or path/data combination.

        Parameters
        ----------
        path
            Filesystem path of this dataset.
        mf
            Manifest file data as dictionary, if data from :path should not be used.
        """
        if not os.path.isdir(path):
            raise EDLError(
                (
                    'Can not load unit from path "{}": Does not specify an ' 'existing directory'
                ).format(path)
            )
        if not mf:
            mf = {}

        path = str(path)
        self._name = os.path.basename(path.rstrip('/\\'))
        self._root_path = os.path.abspath(os.path.join(path, '..'))

        self._attrs = {}
        if not mf:
            with open(os.path.join(self.path, 'manifest.toml'), 'r', encoding='utf-8') as f:
                mf = toml.load(f)

        self._format_version = str(mf.get('format_version', 'unknown'))
        if self._format_version != EDL_FORMAT_VERSION:
            self._root_path = None
            self._name = None
            raise EDLError(
                (
                    'Can not load unit: Format version is unsupported '
                    '(was \'{}\', expected \'{}\').'
                ).format(self._format_version, EDL_FORMAT_VERSION)
            )

        unit_type = mf.get('type')
        if unit_type != self._unit_type:
            self._root_path = None
            self._name = None
            msg = 'EDL Unit type of "{}" can not be loaded by this "{}" object.'.format(
                unit_type, self._unit_type
            )
            raise EDLError(msg)

        if os.path.isfile(os.path.join(self.path, 'attributes.toml')):
            with open(os.path.join(self.path, 'attributes.toml'), 'r', encoding='utf-8') as f:
                self._attrs = toml.load(f)

        raw_time_created = mf['time_created']
        if raw_time_created:
            self._time_created = datetime.fromisoformat(str(raw_time_created))
        if 'collection_id' in mf:
            self._collection_id = uuid.UUID(str(mf['collection_id']))
        if 'generator' in mf:
            self._generator_id = str(mf['generator'])
        if 'authors' in mf:
            self._authors = listify(mf['authors'])

    def _type_as_unittype(self) -> str:
        from .group import EDLGroup
        from .dataset import EDLDataset
        from .collection import EDLCollection

        if isinstance(self, EDLCollection):
            return 'collection'
        if isinstance(self, EDLGroup):
            return 'group'
        if isinstance(self, EDLDataset):
            return 'dataset'
        raise NotImplementedError('An abstract EDL object does not have a defined type.')

    def _make_manifest_dict(self) -> dict[str, T.Any]:
        if not self._time_created:
            # set default creation time, with second-resolution (milliseconds are stripped)
            self._time_created = datetime.now().replace(microsecond=0)

        doc: dict[str, T.Any] = {}
        doc['format_version'] = self._format_version
        doc['type'] = self._type_as_unittype()

        if self.collection_id:
            doc['collection_id'] = str(self.collection_id)

        doc['time_created'] = self._time_created

        if self._generator_id:
            doc['generator'] = self._generator_id

        if self._authors:
            doc['authors'] = self._authors

        return doc

    def _save_metadata(self, manifest: dict[str, T.Any], attributes: dict[str, T.Any]) -> None:
        if not self.path:
            raise EDLError('No path is set for this EDL unit')
        os.makedirs(self.path, exist_ok=True)

        with open(os.path.join(self.path, 'manifest.toml'), 'w', encoding='utf-8') as f:
            toml.dump(manifest, f)

        if attributes:
            with open(os.path.join(self.path, 'attributes.toml'), 'w', encoding='utf-8') as f:
                toml.dump(attributes, f)

    def save(self) -> None:
        raise NotImplementedError('Subclasses need to implement save().')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os
import uuid
import typing as T

import tomlkit as toml

from .unit import EDLUnit, EDLError
from .dataset import EDLDataset


class EDLGroup(EDLUnit):
    """
    An EDL Group
    """

    def __init__(self, name: str | None = None):
        """
        Create a new EDL group.

        If the group has no name and path set, it can not be saved
        to disk.

        Parameters
        ----------
        name
            Name of this group, or None
        """
        EDLUnit.__init__(self, name)
        self._children: list[EDLUnit] = []

    @property
    def children(self) -> list[EDLUnit]:
        return self._children

    @property
    def root_path(self) -> str | None:
        return self._root_path

    @root_path.setter
    def root_path(self, path: str) -> None:
        self._root_path = path
        for c in self._children:
            c.root_path = self.path

    @property
    def collection_id(self) -> uuid.UUID:
        return self._collection_id

    @collection_id.setter
    def collection_id(self, v: uuid.UUID) -> None:
        self._collection_id = v
        for c in self._children:
            c.collection_id = self._collection_id

    def change_name(self, new_name: str) -> None:
        EDLUnit.change_name(self, new_name)
        for c in self._children:
            c.root_path = self.path

    def add_child(self, child: EDLUnit) -> None:
        if not isinstance(child, EDLUnit):
            raise ValueError('Can only have EDL units as children.')
        if not child.name:
            raise ValueError('Child unit must have a name.')
        old_path = None
        if child.root_path:
            old_path = child.path
        child._parent = self
        child.root_path = self.path
        child.collection_id = self.collection_id

        if old_path and os.path.isdir(old_path):
            os.rename(old_path, child.path)
        self._children.append(child)

    @property
    def datasets(self) -> T.Iterator[EDLDataset]:
        for child in self._children:
            if isinstance(child, EDLDataset):
                yield child

    @property
    def groups(self) -> T.Iterator[EDLGroup]:
        for child in self._children:
            if isinstance(child, EDLGroup):
                yield child

    def group_by_name(self, name: str, *, create: bool = False) -> EDLGroup | None:
        group: EDLGroup | None
        for group in self.groups:
            if group.name == name:
                return group

        group = None
        if create:
            group = EDLGroup(name)
            self.add_child(group)
            group.save()

        return group

    def dataset_by_name(self, name: str, *, create: bool = False) -> EDLDataset | None:
        dset: EDLDataset | None
        for dset in self.datasets:
            if dset.name == name:
                return dset

        dset = None
        if create:
            dset = EDLDataset(name)
            self.add_child(dset)
            dset.save()

        return dset

    def save(self) -> None:
        if not self.path:
            raise ValueError('No path set for EDL group "{}"'.format(self.name))
        os.makedirs(self.path, exist_ok=True)

        mf = self._make_manifest_dict()
        self._save_metadata(mf, self.attributes)
        for child in self._children:
            child.save()

    def load(self, path: str | os.PathLike, mf: T.MutableMapping[str, T.Any] | None = None) -> None:
        if not mf:
            mf = {}
        EDLUnit.load(self, path, mf)

        for d in os.listdir(self.path):
            unit_path = os.path.join(self.path, d)
            if not os.path.isdir(unit_path):
                continue
            mf_path = os.path.join(unit_path, 'manifest.toml')
            if not os.path.isfile(mf_path):
                continue

            with open(os.path.join(mf_path), 'r', encoding='utf-8') as f:
                mf = toml.load(f)

            unit_type = mf.get('type')
            unit: EDLGroup | EDLDataset
            if unit_type == 'group':
                unit = EDLGroup()
            elif unit_type == 'dataset':
                unit = EDLDataset()
            else:
                raise EDLError(
                    'EDL unit type "{}" is unknown, data can not be loaded.'.format(unit_type)
                )
            unit.load(unit_path, mf)
            self.add_child(unit)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import typing as T
import importlib

# Determine which loader is responsible for which file type.
DATA_LOADERS = {
    'csv': 'edlio.dataio.csvdata',
    'json': 'edlio.dataio.jsondata',
    'rhd': 'edlio.dataio.intan',
    'tsync': 'edlio.dataio.tsyncfile',
    'video': 'edlio.dataio.video',
}


def load_dataio_module(what: str) -> T.Callable[..., T.Any]:
    path = DATA_LOADERS[what]
    mod = importlib.import_module(path)
    return T.cast(T.Callable[..., T.Any], mod.load_data)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os
import typing as T
import functools

from .unit import EDLUnit, EDLError
from .dataio import DATA_LOADERS, load_dataio_module


@functools.total_ordering
class EDLDataPart:
    """Describes a part of a data block that has been split into multiple files."""

    index: int = -1
    fname: str

    def __init__(self, fname: str, index: int = -1):
        self.fname = fname
        self.index = index

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, EDLDataPart):
            return NotImplemented
        return self.index < other.index

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, EDLDataPart):
            return False
        return self.index == other.index

    def __repr__(self) -> str:
        idx_str = '' if self.index < 0 else ', index=' + (str(self.index))
        return 'EDLDataPart(' + self.fname + idx_str + ')'


class EDLDataFile:
    """A data file, associated with a dataset"""

    parts: list[EDLDataPart] = []

    def __init__(
        self,
        base_path: str | None,
        media_type: str | None = None,
        file_type: str | None = None,
        unit_attrs: dict[str, T.Any] | None = None,
    ):
        if not unit_attrs:
            unit_attrs = {}

        self._base_path = base_path
        self._media_type = media_type
        self._file_type = file_type
        self._unit_attrs = unit_attrs
        self._summary: str | None = None
        self.parts = []

    @property
    def data_type(self) -> tuple[str | None, str | None]:
        return (self._media_type, self._file_type)

    @property
    def media_type(self) -> str | None:
        """The media (MIME) type of this data."""
        return self._media_type

    @media_type.setter
    def media_type(self, mime: str | None) -> None:
        self._media_type = mime

    @property
    def file_type(self) -> str | None:
        """The filetype, in case no media type was available."""
        return self._file_type

    @file_type.setter
    def file_type(self, ftype: str | None) -> None:
        self._file_type = ftype

    @property
    def summary(self) -> str | None:
        """A human-readable summary of what this data is about."""
        return self._summary

    @summary.setter
    def summary(self, text: str | None) -> None:
        """Set the summary text for this data."""
        self._summary = text

    def __repr__(self) -> str:
        data_type = str(self._media_type if self._media_type else self._file_type)
        return (
            'EDLDataFile(type='
            + data_type
            + ', parts='
            + str(self.parts)
            + ', summary='
            + str(self.summary)
            + ')'
        )

    def part_paths(self) -> T.Iterator[str]:
        """
        Return a generator for the path of each file-part, in their correct
        sorting order.
        """
        if not self._base_path:
            raise RuntimeError('Base path for this data file is not set.')
        for part in self.parts:
            yield os.path.join(self._base_path, part.fname)

    def new_part(
        self, fname: str, index: int = -1, *, allow_exists: bool = False
    ) -> tuple[EDLDataPart, str]:
        if not fname:
            raise ValueError('File name is not valid.')
        _, fext = os.path.splitext(fname)
        if fext and fext.startswith('.'):
            fext = fext[1:]
        if not self._file_type:
            self._file_type = fext
        elif self._file_type != fext:
            raise ValueError(
                'New part does not have the right extension for this data: {}'.format(
                    self._file_type
                )
            )

        if not self._base_path:
            raise RuntimeError('Can not add data part: Base path is not set.')

        for ep in self.parts:
            if ep.fname == fname:
                if allow_exists:
                    return ep, os.path.join(self._base_path, ep.fname)
                raise ValueError('A file part with name "{}" already exists.'.format(fname))
        part = EDLDataPart(fname, index)
        self.parts.append(part)
        return part, os.path.join(self._base_path, part.fname)

    def read(
        self, aux_data_entries: T.Optional[T.Sequence[EDLDataFile]] = None, **kwargs: T.Any
    ) -> T.Any:
        """Read all data parts in this set.

        This returns a generator which reads all the individual data parts in this data file.
        The data reader may take auxiliary data into account, if :aux_data is passed.
        """
        if not aux_data_entries:
            aux_data_entries = []

        dclass = self.media_type if self.media_type else self.file_type
        if not dclass:
            raise EDLError(
                (
                    'This data file has no type association. EDL metadata was '
                    'probably invalid, or this file does not exist.'
                )
            )
        if dclass.startswith('video/'):
            dclass = 'video'
        elif dclass.startswith('text/csv'):
            dclass = 'csv'
        elif dclass == 'application/json':
            dclass = 'json'
        elif dclass == 'application/zstd':
            if self.file_type == 'json.zst':
                dclass = 'json'
        elif 'json' in dclass:
            dclass = 'json'

        if dclass not in DATA_LOADERS:
            raise EDLError('I do not know how to read data of type "{}".'.format(dclass))

        if dclass == 'json':
            # knowing the JSON schema in advance is very useful
            kwargs['json_schema'] = self._unit_attrs.get('json_schema')

        load_data = load_dataio_module(dclass)
        return load_data(self.part_paths(), aux_data_entries, **kwargs)


class EDLDataset(EDLUnit):
    """
    An EDL Dataset
    """

    def __init__(self, name: str | None = None):
        """
        Create a new EDL dataset.

        If the dataset has no name and path set, it can not be saved
        to disk.

        Parameters
        ----------
        name
            Name of this dataset, or None
        """
        EDLUnit.__init__(self, name)
        self._data = EDLDataFile(None, unit_attrs=self.attributes)
        self._aux_data: list[EDLDataFile] = []

    @property
    def data(self) -> EDLDataFile:
        return self._data

    @data.setter
    def data(self, df: EDLDataFile) -> None:
        self._data = df

    @property
    def aux_data(self) -> list[EDLDataFile]:
        return self._aux_data

    def add_aux_data(self, adf: EDLDataFile) -> None:
        self._aux_data.append(adf)

    def _parse_data_md(self, d: dict[str, T.Any]) -> EDLDataFile:
        df = EDLDataFile(
            self.path, d.get('media_type'), d.get('file_type'), unit_attrs=self.attributes
        )
        df.summary = d.get('summary')
        for pi in d.get('parts', []):
            df.parts.append(EDLDataPart(pi['fname'], pi.get('index', -1)))
        if df.parts:
            df.parts.sort()
        return df

    def load(
        self, path: str | os.PathLike, mf: T.Optional[T.MutableMapping[str, T.Any]] = None
    ) -> None:
        """
        Load an EDL dataset from a path.

        Parameters
        ----------
        path
            Filesystem path of this dataset.
        mf
            Manifest file data as dictionary, if data from :path should not be used.
        """
        if not mf:
            mf = {}
        EDLUnit.load(self, path, mf)

        self._data = EDLDataFile(self.path, unit_attrs=self.attributes)
        self._aux_data = []
        if 'data' in mf:
            self._data = self._parse_data_md(mf['data'])
        if 'data_aux' in mf:
            daux_raw = mf['data_aux']
            if isinstance(daux_raw, dict):
                self._aux_data.append(self._parse_data_md(daux_raw))
            else:
                for adf_raw in daux_raw:
                    self._aux_data.append(self._parse_data_md(adf_raw))

    def _serialize_data_md(self, df: EDLDataFile) -> dict[str, T.Any]:
        d: dict[str, T.Any] = {}
        if not df.parts:
            return {}
        if self._data.media_type:
            d['media_type'] = df.media_type
        elif self._data.file_type:
            d['file_type'] = df.file_type
        if df.summary:
            d['summary'] = df.summary
        d['parts'] = []
        for part in df.parts:
            pd: dict[str, T.Any] = {'fname': part.fname}
            if part.index >= 0:
                pd['index'] = part.index
            d['parts'].append(pd)
        return d

    def save(self) -> None:
        """Save dataset changes to their current location on disk."""
        if not self.path:
            raise ValueError('No path set for EDL group "{}"'.format(self.name))
        os.makedirs(self.path, exist_ok=True)

        mf = self._make_manifest_dict()
        mf['data'] = self._serialize_data_md(self._data)
        if self._aux_data:
            adf_list = []
            for adf in self._aux_data:
                adf_list.append(self._serialize_data_md(adf))
            mf['data_aux'] = adf_list

        self._save_metadata(mf, self.attributes)

    def read_data(self, **kwargs: T.Any) -> T.Any:
        """Read data from this dataset.

        Returns a generator to read data from this dataset by individual chunks, taking
        auxiliary data into account.
        """
        if not self._data:
            return None
        return self._data.read(self._aux_data, **kwargs)

    def read_aux_data(self, key: str | None = None) -> T.Any | None:
        """
        Read auxiliary data from this dataset.

        Parameters
        ----------
        key
            Identifier key for the auxiliary dataset to load.
            Will look for a substring match in file_type/media_type
            properties of the data first, then look for a summary text match
            to determine which data was requested. If not set, the first dataset
            is loaded.
        Returns
        -------
        The data, or None in case no aux-data or aux-data entry was found.
        """
        if not self._aux_data:
            return None
        if not key:
            return self._aux_data[0].read()
        for adf in self._aux_data:
            if (adf.file_type and key in adf.file_type) or (
                adf.media_type and key in adf.media_type
            ):
                return adf.read()
        for adf in self._aux_data:
            if adf.summary and key in adf.summary:
                return adf.read()
        return None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
from datetime import datetime, timezone

import numpy as np

import edlio
from edlio import ureg
from edlio.dataset import EDLDataFile, EDLDataPart
from edlio.dataio.video import load_data as load_video_data

from . import source_root


def test_load_intan_raw(samples_dir: str) -> None:
    test_coll = edlio.load(os.path.join(samples_dir, 'blink1'))
    assert isinstance(test_coll, edlio.EDLCollection)
    assert test_coll.collection_idname == 'blink1_21-02-08_608bc0ea'

    dset = test_coll.dataset_by_name('intan-signals')

    for intan in dset.read_data(include_nosync_time=True):
        times = intan.sync_times
        dig_chan_all = intan.digin_channels_raw
        assert dig_chan_all.any()

        ts_raw = intan._raw_data['timestamp']
        assert ts_raw.any()
        assert intan._timestamp_len == 1304640
        assert intan.digin_channels_raw[0].size == 1304640

        assert times[0] == 6.196 * ureg.ms
        assert intan._nosync_ts[0] == 6.196 * ureg.ms


def test_load_tsync_only(samples_dir: str) -> None:
    test_coll = edlio.load(os.path.join(samples_dir, 'blink1'))
    assert isinstance(test_coll, edlio.EDLCollection)
    assert test_coll.collection_idname == 'blink1_21-02-08_608bc0ea'

    dset = test_coll.group_by_name('videos').dataset_by_name('miniscope')

    tsync_data = [tsync for tsync in dset.read_aux_data('tsync')]
    assert len(tsync_data) == 1
    tsync = tsync_data[0]

    assert tsync.time_units == (ureg.dimensionless, ureg.millisecond)
    assert tsync.time_labels == ('frame-no', 'master-time')
    assert tsync.time_created == datetime.fromisoformat('2021-02-08 17:17:31').replace(
        tzinfo=timezone.utc
    )
    assert tsync.times.shape == (1287, 2)
    assert tsync.times.size == 2574


def test_load_video_tsync_preserves_millisecond_timestamp_scale() -> None:
    dataset_path = os.path.join(
        source_root, 'tests', 'samples', 'blink1', 'videos', 'generic-camera'
    )

    aux_data = EDLDataFile(dataset_path, file_type='tsync')
    aux_data.parts.append(EDLDataPart('video_timestamps.tsync', 0))
    tsync = next(aux_data.read())

    frames = load_video_data([os.path.join(dataset_path, 'video.mkv')], [aux_data])
    frame = next(frames)
    expected_time_msec = tsync.times[0, 1]
    assert frame.time.to(tsync.time_units[1]).magnitude == expected_time_msec


def test_load_video_tsync_keeps_frame_index_unitless() -> None:
    dataset_path = os.path.join(
        source_root, 'tests', 'samples', 'blink1', 'videos', 'generic-camera'
    )

    aux_data = EDLDataFile(dataset_path, file_type='tsync')
    aux_data.parts.append(EDLDataPart('video_timestamps.tsync', 0))
    tsync = next(aux_data.read())

    frame = next(load_video_data([os.path.join(dataset_path, 'video.mkv')], [aux_data]))
    assert frame.index == tsync.times[0, 0]
    assert not hasattr(frame.index, 'units')


def test_load_json_csv(samples_dir: str) -> None:
    jcstore = edlio.load(os.path.join(samples_dir, 'jsoncsv1'))
    assert isinstance(jcstore, edlio.EDLCollection)
    assert jcstore.collection_idname == 'jsoncsv1_24-02-18_1779a25f'

    # check reading a pandas-extended JSON integer numbers set
    dset = jcstore.dataset_by_name('numbers-json')
    assert dset.attributes == {
        'json_data_unit': 'au',
        'json_schema': 'extended-pandas',
        'json_time_unit': 'milliseconds',
    }
    dfs = list(dset.read_data())
    assert len(dfs) == 1
    assert dfs[0].dtypes.array == [np.int64, np.int64]
    assert dfs[0].columns.tolist() == ['timestamp_msec', 'Int 1']
    assert dfs[0].shape == (4524, 2)

    # check reading a pandas-compatible JSON float numbers set
    dset = jcstore.dataset_by_name('sines-json')
    assert dset.attributes == {
        'json_data_unit': 'au',
        'json_schema': 'pandas-split',
        'json_time_unit': 'milliseconds',
    }
    dfs = list(dset.read_data())
    assert len(dfs) == 1
    assert dfs[0].dtypes.array == [np.int64, np.float64, np.float64, np.float64]
    assert dfs[0].columns.tolist() == ['timestamp_msec', 'Sine 1', 'Sine 2', 'Sine 3']
    assert dfs[0].shape == (4524, 4)

    # check reading a JSON string table
    dset = jcstore.dataset_by_name('table-json')
    assert dset.attributes == {'json_schema': 'pandas-split'}
    dfs = list(dset.read_data())
    assert len(dfs) == 1
    assert dfs[0].dtypes.array == [np.int64, np.dtype('O'), np.dtype('O')]
    assert dfs[0].columns.tolist() == ['Time', 'Tag', 'Value']
    assert dfs[0].shape == (5, 3)

    # check reading a CSV table as Pandas DataFrame
    dset = jcstore.dataset_by_name('table-csv')
    dfs = list(dset.read_data(as_dataframe=True))
    assert len(dfs) == 1
    assert dfs[0].dtypes.array == [np.int64, np.dtype('O'), np.dtype('O')]
    assert dfs[0].columns.tolist() == ['Time', 'Tag', 'Value']
    assert dfs[0].shape == (5, 3)

    # check reading a CSV table row-by-row
    dset = jcstore.dataset_by_name('table-csv')
    rows = list(dset.read_data())
    assert len(rows) == 6
    assert rows[0] == ['Time', 'Tag', 'Value']
    assert rows[-1] == ['20015', 'beta', 'eLcwGIFVu1A9NV']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from .group import EDLGroup


class EDLCollection(EDLGroup):
    """
    An EDL Collection
    """

    def __init__(self, name: str | None = None):
        """
        Create a new EDL collection.

        If the collection has no name and path set, it can not be saved
        to disk.

        Parameters
        ----------
        name
            Name of this collection, or None
        """
        EDLGroup.__init__(self, name)

    @property
    def generator_id(self) -> str | None:
        """Identification string of the software which generated this EDL unit."""
        return self._generator_id

    @generator_id.setter
    def generator_id(self, v: str) -> None:
        self._generator_id = v

    @property
    def collection_idname(self) -> str:
        """
        Retrieve a human-readable string for this collection which is
        most likely (but not guaranteed to be) unique.
        If this collection has no properties set, the value of this property
        may be None or empty.
        The format of the returned string is arbitrary and should not be parsed.
        """
        parts = []
        subject_id = self.attributes.get('subject_id')
        if subject_id:
            parts.append(str(subject_id))
        if self.name:
            parts.append(str(self.name))
        if self.time_created:
            parts.append(self.time_created.strftime('%y-%m-%d'))
        if self.collection_id:
            parts.append(str(self.collection_id)[-8:])

        return '_'.join(parts).replace(' ', '')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import csv
import typing as T

if T.TYPE_CHECKING:
    import pandas as pd


def load_data(
    part_paths: T.Iterable[str], aux_data_entries: T.Any, as_dataframe: bool = False
) -> T.Iterator[pd.DataFrame | list[str]]:
    """Entry point for automatic dataset loading.

    This function is used internally to load CSV data.
    """

    if as_dataframe:
        import pandas as pd

        for fname in part_paths:
            df = pd.read_csv(fname, sep=';')
            yield df
    else:
        for fname in part_paths:
            with open(fname, newline='', encoding='utf-8') as f:
                reader = csv.reader(f, delimiter=';')
                for row in reader:
                    yield row



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

"""
Module to work with data in an Experiment Directory Layout (EDL) structure
"""

from __future__ import annotations

__appname__ = 'edlio'
__version__ = '0.3.0'

import os

import pint
import tomlkit as toml

from .unit import EDLError
from .group import EDLGroup
from .dataset import EDLDataset
from .collection import EDLCollection

__all__ = ['ureg', 'Q_', 'EDLError', 'EDLGroup', 'EDLCollection', 'EDLDataset', 'load']


# Pint default registry for unit conversions
ureg = pint.get_application_registry()
Q_ = ureg.Quantity


def load(path: str | os.PathLike[str]) -> EDLCollection | EDLGroup | EDLDataset:
    """
    Open an EDL unit via its filesystem path.

    This function will read an EDL unit on the filesystem and return
    an object representing it.
    Depending on the type of the EDL unit, the returned datatype may be
    a collection, group or dataset.

    Parameters
    ----------
    path
        The filesystem location of the EDL unit.

    Returns
    -------
    EDLCollection, EDLGroup, EDLDataset
        An EDL unit.
    """

    mf_path = os.path.join(path, 'manifest.toml')
    if not os.path.isfile(mf_path):
        raise EDLError('The directory "{}" is no valid EDL unit.'.format(path))

    with open(os.path.join(mf_path), 'r', encoding='utf-8') as f:
        mf = toml.load(f)

    unit_type = mf.get('type')
    unit: EDLCollection | EDLGroup | EDLDataset
    if unit_type == 'collection':
        unit = EDLCollection()
    elif unit_type == 'group':
        unit = EDLGroup()
    elif unit_type == 'dataset':
        unit = EDLDataset()
    else:
        raise EDLError('EDL unit type "{}" is unknown, data can not be loaded.'.format(unit_type))

    unit.load(path, mf)
    return unit



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import typing as T
import logging as log

import pint
import numpy as np

try:
    from neo.rawio import IntanRawIO
    from neo.io.basefromrawio import BaseFromRaw
except ImportError as e:
    raise ImportError(
        'Unable to find the neo module. Can not read Intan electrophysiology data. {}'.format(
            str(e)
        )
    ) from e

from .. import ureg
from ..dataset import EDLDataFile
from .tsyncfile import TSyncFileMode


# pylint: disable=abstract-method
class SyncIntanReader(IntanRawIO, BaseFromRaw):
    """Reader for Intan electrophysiology data.

    This class is the same as IntanIO from NEO, but additionally also provides
    synchronized timestamps in its :meth:`sync_times` property.
    """

    _prefered_signal_group_mode = 'group-by-same-units'

    def __init__(self, intan_filename: str):
        IntanRawIO.__init__(self, filename=intan_filename)
        BaseFromRaw.__init__(self, intan_filename)
        self._sync_ts = None
        self._timestamp_len = 0
        self._digin_channels: np.ndarray | None = None

    def _parse_header(self) -> None:
        IntanRawIO._parse_header(self)

    @property
    def has_adjusted_times(self) -> bool:
        """Returns True if we have synchronized or otherwise adjusted time data."""
        return self._sync_ts is not None

    @property
    def sync_times(self) -> T.Any:
        """Synchronized timestamps vector"""
        return self._sync_ts

    @property
    def digin_channels_raw(self) -> np.ndarray:
        """Obtain the raw data of digital input channels"""
        if self._digin_channels is not None:
            return self._digin_channels

        ts_len = self._timestamp_len
        digin_raw = None
        dig_fields = ('DIGITAL-IN', 'USB board digital input channel')
        for field in dig_fields:
            if field in self._raw_data.dtype.names:  # pylint: disable=no-member
                digin_raw = self._raw_data[field].flatten()
                break
        if digin_raw is None:
            raise ValueError(
                'Unable to find board digital input channel data! (Available fields: {})'.format(
                    self._raw_data.dtype.names  # pylint: disable=no-member
                )
            )

        if digin_raw.size > 0:
            # FIXME: We just assume 16 digital channels. This code should actually go into NEO
            # in a better form, instead of being hacked in here
            digin_chan_n = 16
            self._digin_channels = np.zeros((digin_chan_n, ts_len), dtype=bool)
            for i in range(0, digin_chan_n):
                self._digin_channels[i, :] = np.not_equal(np.bitwise_and(digin_raw, (1 << i)), 0)
        else:
            # empty dummy value, so we don't try to load this again
            self._digin_channels = np.zeros((0, 0), dtype=bool)

        return self._digin_channels


def _make_nosync_tsvec(
    data_len: int, sample_rate: pint.Quantity[int], init_offset: pint.Quantity[int]
) -> pint.Quantity[np.ndarray]:
    """Create time vector taking only the start time into account."""
    tv = np.arange(0, data_len).astype(np.float64) * ureg.dimensionless
    tv = (tv / sample_rate).to(ureg.msec) - init_offset.to(ureg.msec)

    return tv


def _make_synced_tsvec(
    data_len: int,
    sample_rate: pint.Quantity[int],
    idx_intan: np.ndarray,
    sync_map: pint.Quantity[np.ndarray],
) -> pint.Quantity[np.ndarray]:
    """Create time vector, synchronizing all timepoints.

    Syntalos monitors the timestamps coming from the external device and compares them to a
    prediction about what the actual timestamp based on its master clock should be.
    If the device timestamp is too large or to small compared to the prediction,
    Syntalos writes down the timepoint on its master clock next to the timestamp of the external
    device.
    Then it resets the prediction again based on the newly corrected time.

    To get continuous synchronized timestamps from the data provided by Syntalos, we just do a
    linear interpolation from one time point to the next.
    To do this quickly, we calculate the sampling index (based on the device time and device
    sampling rate) of each timepoint in :sync_map and move from one to the other.

    This works very well with accurate timing devices like Intan's RHD2000, CED, etc.
    It has *not* been tested with every device running its own clock.

    This specific function is written only for use with Intan (as it makes certain assumptions
    about the structure of the data).
    """

    sync_len = sync_map.shape[0]
    if sync_len <= 2:
        # nothing to synchronize (we just have the initial offset and the end point),
        # just return the shifted vector
        log.debug(
            'Intan time sync map was too short for synchronization, '
            'returning timeshifted timestamp vector.'
        )
        init_offset = sync_map[0][0] - sync_map[0][1]
        return _make_nosync_tsvec(data_len, sample_rate, init_offset)

    tv_adj = np.zeros((data_len,), dtype=np.float64) * ureg.msec
    for i in range(sync_len - 1):
        # beginning of the 'frame' in master clock time
        m_start = sync_map[i, 1].to(ureg.msec)
        # end of the 'frame' in master clock time
        m_end = sync_map[i + 1, 1].to(ureg.msec)

        # beginning of the 'frame' in Intan sample indices
        d_start = idx_intan[i]
        # end of the 'frame' in Intan sample indices
        d_end = idx_intan[i + 1] + 1
        if d_end > data_len:
            log.error(
                'Intan sync index is bigger than the amount of recorded data ({} > {}). '
                'This means data may be missing or the time-sync files does not belong '
                'to this dataset'.format(d_end, data_len)
            )
            # try to do something semi-sensible, then abort as there is nothing we can do anymore
            d_end = data_len
            tv_adj[d_start:d_end] = np.linspace(m_start, m_end, d_end - d_start)
            break

        # linear interpolation between the beginning time point and the end time point
        # from the beginning sample index up to the end sample index
        tv_adj[d_start:d_end] = np.linspace(m_start, m_end, d_end - d_start)

        # last timepoint: just use the same slope and extrapolate to the end
        if i == (sync_len - 2):
            slope_part = (m_end - m_start) / (d_end - d_start)
            values_part = (np.arange(d_end, data_len) - d_end) * slope_part
            tv_adj[d_end : data_len + 1] = values_part + tv_adj[d_end - 1]
            break

    return tv_adj


def load_data(
    part_paths: T.Iterable[str],
    aux_data_entries: T.Sequence[EDLDataFile],
    do_timesync: bool = True,
    include_nosync_time: bool = False,
) -> T.Iterator[SyncIntanReader]:
    """Entry point for automatic dataset loading.

    This function is used internally to load Intan RHD signals data
    and apply time synchronization.
    """

    start_offset = 0 * ureg.usec
    sync_map = np.empty([0, 2]) * ureg.dimensionless

    aux_data: EDLDataFile | None = None
    for adf in aux_data_entries:
        if (adf.file_type and 'tsync' in adf.file_type) or (
            adf.media_type and 'tsync' in adf.media_type
        ):
            aux_data = adf
            break

    if aux_data:
        tsf_count = 0
        for tsf in aux_data.read():
            if tsf.sync_mode != TSyncFileMode.SYNCPOINTS:
                raise ValueError(
                    'Can not synchronize RHD signal timestamps using a tsync file '
                    'that is not in \'syncpoints\' mode.'
                )
            if tsf.time_units != (ureg.usec, ureg.usec):
                raise ValueError(
                    'For RHD signal synchronization, both timestamp units in tsync '
                    'file must be microseconds. Found: {}'.format(tsf.time_units)
                )
            sync_map = np.vstack((sync_map, tsf.times)) * ureg.usec
            tsf_count += 1
        if tsf_count > 1:
            log.warning(
                'More than one tsync file found for Intan data '
                '- this is unusual and not a well-tested scenario.'
            )

        # the very first entry in the tsync file is the initial Intan to master-clock offset
        start_offset = sync_map[0][0] - sync_map[0][1]

    has_sync_info = sync_map.size > 0
    log.info(
        'Initial RHD time offset: {} µs ({})'.format(
            start_offset.magnitude, 'sync info found' if has_sync_info else 'no sync info'
        )
    )

    intan_readers = []
    for fname in part_paths:
        reader = SyncIntanReader(fname)
        if not has_sync_info:
            yield reader
            continue
        # data is lazily loaded from an mmap'ed file, so we can stash all readers here for now
        intan_readers.append(reader)

    # sanity checks and collect absolute data length of the whole recording
    recording_data_len = 0
    sample_rate = None
    for reader in intan_readers:
        if sample_rate is None:
            sample_rate = reader._max_sampling_rate * ureg.hertz
        else:
            if sample_rate != reader._max_sampling_rate * ureg.hertz:
                raise ValueError(
                    'Samplig rate in Intan recording slice file differs from previous '
                    'files. The data may not belong to the same recording.'
                )
        reader._timestamp_len = reader._raw_data['timestamp'].size
        recording_data_len += reader._timestamp_len

    if sample_rate is None:
        raise ValueError('Unable to determine sample rate from Intan data.')

    # convert to seconds and multiply with sampling rate to obtain sample indices
    intan_sync_idx = (sync_map[:, 0].to(ureg.seconds) * sample_rate).magnitude
    intan_sync_idx = intan_sync_idx.astype(np.int32)

    tvec_noadj: pint.Quantity[np.ndarray] | None = None  # silence pylint
    if include_nosync_time:
        tvec_noadj = _make_nosync_tsvec(recording_data_len, sample_rate, start_offset)
    if do_timesync:
        tvec = _make_synced_tsvec(recording_data_len, sample_rate, intan_sync_idx, sync_map)
    else:
        tvec = (
            tvec_noadj
            if tvec_noadj is not None
            else _make_nosync_tsvec(recording_data_len, sample_rate, start_offset)
        )

    last_ts_idx = 0
    for reader in intan_readers:
        ts_len = reader._timestamp_len

        if tvec_noadj is not None:
            reader._nosync_ts = tvec_noadj[last_ts_idx : last_ts_idx + ts_len]

        reader._sync_ts = tvec[last_ts_idx : last_ts_idx + ts_len]
        last_ts_idx += ts_len

        yield reader



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os
import json
import typing as T

if T.TYPE_CHECKING:
    import pandas as pd


def _read_pandas_extended_json(fname: str | os.PathLike) -> pd.DataFrame:
    import pandas as pd

    if str(fname).endswith('.zst'):
        try:
            import zstandard as zstd
        except ImportError as e:
            raise ImportError(
                'Missing optional dependency "zstandard". Please install it with pip!'
            ) from e

        with open(fname, 'rb') as f:
            dctx = zstd.ZstdDecompressor()
            zr = dctx.stream_reader(f)
            jd = json.load(zr)
    else:
        with open(fname, 'r', encoding='utf-8') as f:
            jd = json.load(f)

    jd.pop('collection_id')
    jd.pop('time_unit')
    jd.pop('data_unit')
    df = pd.DataFrame(jd.pop('data'), columns=jd.pop('columns'))

    return df


def load_data(
    part_paths: T.Iterable[str], aux_data_entries: T.Any, json_schema: str | None = None
) -> T.Iterator[pd.DataFrame]:
    """Entry point for automatic dataset loading.

    This function is used internally to load JSON data.
    """
    try:
        import pandas as pd
    except ImportError as e:
        raise ImportError(
            'Missing optional dependency "pandas". Please install it with pip!'
        ) from e

    if json_schema == 'extended-pandas':
        for fname in part_paths:
            yield _read_pandas_extended_json(fname)
    else:
        for fname in part_paths:
            df = pd.read_json(fname, orient='split')
            yield df



#!/usr/bin/env python3

import os
import pathlib

try:
    import tomllib as toml
except ModuleNotFoundError:
    import tomlkit as toml

from setuptools import setup

from edlio import __appname__, __version__

source_root = os.path.dirname(__file__)
if not os.path.isabs(__file__):
    thisfile = os.path.dirname(os.path.normpath(os.path.join(os.getcwd(), __file__)))

packages = [
    'edlio',
    'edlio.dataio',
]

pyproject_data = toml.loads(pathlib.Path('pyproject.toml').read_text(encoding='utf-8'))

setup(
    name=__appname__,
    version=__version__,
    author="Matthias Klumpp",
    author_email="matthias@tenstral.net",
    description='Module to work with data in an Experiment Directory Layout (EDL) structure',
    url="https://edl.readthedocs.io/",
    license=pyproject_data['project']['license']['text'],
    long_description=open(os.path.join(source_root, 'README.md'), encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    install_requires=pyproject_data['project']['dependencies'],
    tests_require=pyproject_data['project']['optional-dependencies']['test'],
    python_requires=pyproject_data['project']['requires-python'],
    platforms=['any'],
    packages=packages,
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import random
import string
import typing as T
import collections

_T = T.TypeVar('_T')


def sanitize_name(name: str | None) -> str | None:
    """
    Sanitize a string for use as an EDL unit name,
    by stripping or replacing invalid characters.

    Parameters
    ----------
    name
        A string to sanitize.

    Returns
    -------
    str
        The sanitized name.
    """
    if not name:
        return None
    s = ''.join(filter(lambda x: x in string.printable, name))
    s = s.replace('/', '_').replace('\\', '_').replace(':', '')
    # TODO: Replace Windows' reserved names
    if not s:
        s = ''.join(random.choices(string.ascii_uppercase + string.digits, k=4))
    return s


@T.overload
def listify(item: list[_T]) -> list[_T]: ...


@T.overload
def listify(item: _T) -> list[_T]: ...


def listify(item: T.Any) -> list[T.Any]:
    """
    Return a list of :item, unless :item already is a list.
    """
    if not item:
        return []
    if type(item) is list:
        return item
    if isinstance(item, collections.abc.Sequence):
        return list(item)
    return [item]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2026 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import typing as T
from collections.abc import Generator

import cv2 as cv
import numpy as np
from pint.facets.plain import PlainQuantity

from .. import ureg
from ..dataset import EDLDataFile
from .tsyncfile import TSyncFileMode


class Frame:
    mat: np.ndarray
    time: int | PlainQuantity[int]
    index: int

    def __init__(self, mat: np.ndarray, time: int | PlainQuantity[int], index: int):
        """Create a new frame representation for a video file.

        Parameters
        ----------
        mat
            OpenCV matrix containing the image data.
        time
            Time as quantity (usually in microseconds)
        index
            An optional frame index that increases monotonically.
        """
        self.mat = mat
        self.time = time
        self.index = index


def _read_video_aux_data(
    aux_data: EDLDataFile,
) -> Generator[tuple[int, PlainQuantity[int]], None, None]:
    if aux_data.file_type == 'csv' or aux_data.media_type == 'text/csv':
        for index, timestamp in aux_data.read():
            if index == 'frame':
                # we have a table header, skip it
                continue
            yield int(index), int(timestamp) * ureg.usec
        return

    if aux_data.file_type == 'tsync':
        for tsf in aux_data.read():
            if tsf.sync_mode != TSyncFileMode.CONTINUOUS:
                raise ValueError(
                    'Can not synchronize video timestamps using a non-continuous tsync file.'
                )
            if tsf.time_units[0] != ureg.dimensionless:
                raise ValueError(
                    'Unit of first time in tsync mapping has to be \'index\' for video files.'
                )
            if tsf.time_units[1] not in (ureg.msec, ureg.usec):
                raise ValueError(
                    'We currently expect video timestamps to be in microseconds or milliseconds (unit was {}).'.format(
                        tsf.time_units[1]
                    )
                )
            for index, timestamp in tsf.times:
                yield int(index), (int(timestamp) * tsf.time_units[1]).to(ureg.usec)

        return

    raise ValueError(
        'Unknown auxiliary data type ({}|{}) for video file.'.format(
            aux_data.file_type, aux_data.media_type
        )
    )


def load_data(
    part_paths: T.Iterable[str], aux_data_entries: T.Sequence[EDLDataFile]
) -> T.Iterator[Frame]:
    """Entry point for automatic dataset loading.

    This function is used internally to load data from a video and expose
    it as stream of frames.
    """
    aux_data: EDLDataFile | None = None
    valid_timestamp_aux_keys = ['tsync', 'csv']
    for adf in aux_data_entries:
        for vtak in valid_timestamp_aux_keys:
            if (adf.file_type and vtak in adf.file_type) or (
                adf.media_type and vtak in adf.media_type
            ):
                aux_data = adf
                break
        if aux_data is not None:
            break

    sync_map_gen = None
    if aux_data:
        sync_map_gen = _read_video_aux_data(aux_data)

    frame_index = 0
    for fname in part_paths:
        vc = cv.VideoCapture(fname)
        while True:
            ret, mat = vc.read()
            if not ret:
                break
            if sync_map_gen is None:
                frame = Frame(mat, time=-1, index=frame_index)
            else:
                index, time = next(sync_map_gen)
                frame = Frame(mat, time=time, index=index)
            yield frame
            frame_index += 1
